# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import six

from keystoneauth1 import exceptions
from mistralclient.api import base
from mistralclient import utils


urlparse = six.moves.urllib.parse


class Workflow(base.Resource):
    resource_name = 'Workflow'


class WorkflowManager(base.ResourceManager):
    resource_class = Workflow

    def create(self, definition, namespace='', scope='private'):
        self._ensure_not_empty(definition=definition)

        # If the specified definition is actually a file, read in the
        # definition file
        definition = utils.get_contents_if_file(definition)

        try:
            resp = self.http_client.post(
                '/workflows?scope=%s&namespace=%s' % (scope, namespace),
                definition,
                headers={'content-type': 'text/plain'}
            )
        except exceptions.HttpError as ex:
            self._raise_api_exception(ex.response)

        if resp.status_code != 201:
            self._raise_api_exception(resp)

        return [self.resource_class(self, resource_data)
                for resource_data in base.extract_json(resp, 'workflows')]

    def update(self, definition, namespace='', scope='private', id=None):
        self._ensure_not_empty(definition=definition)

        url_pre = ('/workflows/%s' % id) if id else '/workflows'

        # If the specified definition is actually a file, read in the
        # definition file
        definition = utils.get_contents_if_file(definition)

        try:
            resp = self.http_client.put(
                '%s?namespace=%s&scope=%s' % (url_pre, namespace, scope),
                definition,
                headers={'content-type': 'text/plain'}
            )
        except exceptions.HttpError as ex:
            self._raise_api_exception(ex.response)

        if resp.status_code != 200:
            self._raise_api_exception(resp)

        if id:
            return self.resource_class(self, base.extract_json(resp, None))

        return [self.resource_class(self, resource_data)
                for resource_data in base.extract_json(resp, 'workflows')]

    def list(self, namespace='', marker='', limit=None, sort_keys='',
             sort_dirs='', **filters):
        qparams = {}

        if namespace:
            qparams['namespace'] = namespace

        if marker:
            qparams['marker'] = marker

        if limit and limit > 0:
            qparams['limit'] = limit

        if sort_keys:
            qparams['sort_keys'] = sort_keys

        if sort_dirs:
            qparams['sort_dirs'] = sort_dirs

        for name, val in filters.items():
            qparams[name] = val

        query_string = ("?%s" % urlparse.urlencode(list(qparams.items()))
                        if qparams else "")

        return self._list(
            '/workflows%s' % query_string,
            response_key='workflows',
        )

    def get(self, identifier, namespace=''):
        self._ensure_not_empty(identifier=identifier)

        return self._get(
            '/workflows/%s?namespace=%s' % (identifier, namespace)
        )

    def delete(self, identifier, namespace=None):
        self._ensure_not_empty(identifier=identifier)

        path = '/workflows/%s' % identifier

        if namespace:
            path = path + '?namespace=%s' % namespace

        self._delete(path)

    def validate(self, definition):
        self._ensure_not_empty(definition=definition)

        # If the specified definition is actually a file, read in the
        # definition file
        definition = utils.get_contents_if_file(definition)

        try:
            resp = self.http_client.post(
                '/workflows/validate',
                definition,
                headers={'content-type': 'text/plain'}
            )
        except exceptions.HttpError as ex:
            self._raise_api_exception(ex.response)

        if resp.status_code != 200:
            self._raise_api_exception(resp)

        return base.extract_json(resp, None)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os.path
import pkg_resources as pkg
from six.moves.urllib import parse
from six.moves.urllib import request

from mistralclient.api import base as api_base
from mistralclient.api.v2 import workbooks
from mistralclient.tests.unit.v2 import base

# TODO(everyone): later we need additional tests verifying all the errors etc.


WB_DEF = """
---
version: 2.0

name: wb

workflows:
  wf1:
    type: direct
    input:
      - param1
      - param2

    tasks:
      task1:
        action: std.http url="localhost:8989"
        on-success:
          - test_subsequent

      test_subsequent:
        action: std.http url="http://some_url" server_id=1
"""

INVALID_WB_DEF = """
version: 2.0

name: wb

workflows:
  wf1:
    type: direct
    tasks:
      task1:
        action: std.http url="localhost:8989"
        workflow: wf2
"""

WORKBOOK = {'definition': WB_DEF}

URL_TEMPLATE = '/workbooks'
URL_TEMPLATE_NAME = '/workbooks/%s'
URL_TEMPLATE_VALIDATE = '/workbooks/validate'


class TestWorkbooksV2(base.BaseClientV2Test):
    def test_create(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE,
                                json=WORKBOOK,
                                status_code=201)

        wb = self.workbooks.create(WB_DEF)

        self.assertIsNotNone(wb)
        self.assertEqual(WB_DEF, wb.definition)

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_create_with_file_uri(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE,
                                json=WORKBOOK,
                                status_code=201)

        # The contents of wb_v2.yaml must be identical to WB_DEF
        path = pkg.resource_filename(
            'mistralclient',
            'tests/unit/resources/wb_v2.yaml'
        )
        path = os.path.abspath(path)

        # Convert the file path to file URI
        uri = parse.urljoin('file:', request.pathname2url(path))

        wb = self.workbooks.create(uri)

        self.assertIsNotNone(wb)
        self.assertEqual(WB_DEF, wb.definition)

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_update(self):
        self.requests_mock.put(self.TEST_URL + URL_TEMPLATE, json=WORKBOOK)

        wb = self.workbooks.update(WB_DEF)

        self.assertIsNotNone(wb)
        self.assertEqual(WB_DEF, wb.definition)

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_update_with_file(self):
        self.requests_mock.put(self.TEST_URL + URL_TEMPLATE, json=WORKBOOK)

        # The contents of wb_v2.yaml must be identical to WB_DEF
        path = pkg.resource_filename(
            'mistralclient',
            'tests/unit/resources/wb_v2.yaml'
        )

        wb = self.workbooks.update(path)

        self.assertIsNotNone(wb)
        self.assertEqual(WB_DEF, wb.definition)

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_list(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE,
                               json={'workbooks': [WORKBOOK]})

        workbook_list = self.workbooks.list()

        self.assertEqual(1, len(workbook_list))

        wb = workbook_list[0]

        self.assertEqual(
            workbooks.Workbook(self.workbooks, WORKBOOK).to_dict(),
            wb.to_dict()
        )

    def test_get(self):
        url = self.TEST_URL + URL_TEMPLATE_NAME % 'wb'
        self.requests_mock.get(url, json=WORKBOOK)

        wb = self.workbooks.get('wb')

        self.assertIsNotNone(wb)
        self.assertEqual(
            workbooks.Workbook(self.workbooks, WORKBOOK).to_dict(),
            wb.to_dict()
        )

    def test_delete(self):
        url = self.TEST_URL + URL_TEMPLATE_NAME % 'wb'
        self.requests_mock.delete(url, status_code=204)

        self.workbooks.delete('wb')

    def test_validate(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE_VALIDATE,
                                json={'valid': True})

        result = self.workbooks.validate(WB_DEF)

        self.assertIsNotNone(result)
        self.assertIn('valid', result)
        self.assertTrue(result['valid'])

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_validate_with_file(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE_VALIDATE,
                                json={'valid': True})

        # The contents of wb_v2.yaml must be identical to WB_DEF
        path = pkg.resource_filename(
            'mistralclient',
            'tests/unit/resources/wb_v2.yaml'
        )

        result = self.workbooks.validate(path)

        self.assertIsNotNone(result)
        self.assertIn('valid', result)
        self.assertTrue(result['valid'])

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_validate_failed(self):
        mock_result = {
            "valid": False,
            "error": "Task properties 'action' and 'workflow' "
                     "can't be specified both"
        }

        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE_VALIDATE,
                                json=mock_result)

        result = self.workbooks.validate(INVALID_WB_DEF)

        self.assertIsNotNone(result)
        self.assertIn('valid', result)
        self.assertFalse(result['valid'])
        self.assertIn('error', result)
        self.assertIn(
            "Task properties 'action' and 'workflow' "
            "can't be specified both", result['error']
        )

        last_request = self.requests_mock.last_request

        self.assertEqual(INVALID_WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])

    def test_validate_api_failed(self):
        self.requests_mock.post(self.TEST_URL + URL_TEMPLATE_VALIDATE,
                                status_code=500)

        self.assertRaises(
            api_base.APIException,
            self.workbooks.validate,
            WB_DEF
        )

        last_request = self.requests_mock.last_request

        self.assertEqual(WB_DEF, last_request.text)
        self.assertEqual('text/plain', last_request.headers['content-type'])



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse

from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils


def format(workbook=None):
    columns = (
        'Name',
        'Tags',
        'Scope',
        'Created at',
        'Updated at'
    )

    if workbook:
        data = (
            workbook.name,
            base.wrap(', '.join(workbook.tags or '')) or '<none>',
            workbook.scope,
            workbook.created_at,
        )

        if hasattr(workbook, 'updated_at'):
            data += (workbook.updated_at,)
        else:
            data += (None,)

    else:
        data = (tuple('' for _ in range(len(columns))),)

    return columns, data


class List(base.MistralLister):
    """List all workbooks."""

    def _get_format_function(self):
        return format

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        return mistral_client.workbooks.list()


class Get(command.ShowOne):
    """Show specific workbook."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument(
            'workbook',
            help='Workbook name'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        workbook = mistral_client.workbooks.get(parsed_args.workbook)

        return format(workbook)


class Create(command.ShowOne):
    """Create new workbook."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workbook definition file'
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag workbook will be marked as "public".'
        )

        return parser

    def take_action(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine
        workbook = mistral_client.workbooks.create(
            parsed_args.definition.read(),
            scope=scope
        )

        return format(workbook)


class Delete(command.Command):
    """Delete workbook."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument('workbook', nargs='+', help='Name of workbook(s).')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        utils.do_action_on_many(
            lambda s: mistral_client.workbooks.delete(s),
            parsed_args.workbook,
            "Request to delete workbook %s has been accepted.",
            "Unable to delete the specified workbook(s)."
        )


class Update(command.ShowOne):
    """Update workbook."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workbook definition file'
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag workbook will be marked as "public".'
        )

        return parser

    def take_action(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine
        workbook = mistral_client.workbooks.update(
            parsed_args.definition.read(),
            scope=scope
        )

        return format(workbook)


class GetDefinition(command.Command):
    """Show workbook definition."""

    def get_parser(self, prog_name):
        parser = super(GetDefinition, self).get_parser(prog_name)

        parser.add_argument('name', help='Workbook name')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        definition = mistral_client.workbooks.get(parsed_args.name).definition

        self.app.stdout.write(definition or "\n")


class Validate(command.ShowOne):
    """Validate workbook."""

    def _format(self, result=None):
        columns = ('Valid', 'Error')

        if result:
            data = (result.get('valid'), result.get('error'),)
        else:
            data = (tuple('<none>' for _ in range(len(columns))),)

        return columns, data

    def get_parser(self, prog_name):
        parser = super(Validate, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Workbook definition file'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        result = mistral_client.workbooks.validate(
            parsed_args.definition.read()
        )

        return self._format(result)



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Command-line interface to the Mistral APIs
"""

import argparse
import logging
import os
import sys

from cliff import app
from cliff import commandmanager
from osc_lib.command import command

from mistralclient.api import client
from mistralclient.auth import auth_types
import mistralclient.commands.v2.action_executions
import mistralclient.commands.v2.actions
import mistralclient.commands.v2.cron_triggers
import mistralclient.commands.v2.environments
import mistralclient.commands.v2.event_triggers
import mistralclient.commands.v2.executions
import mistralclient.commands.v2.members
import mistralclient.commands.v2.services
import mistralclient.commands.v2.tasks
import mistralclient.commands.v2.workbooks
import mistralclient.commands.v2.workflows
from mistralclient import exceptions as exe


def env(*args, **kwargs):
    """Returns the first environment variable set.

    If all are empty, defaults to '' or keyword arg `default`.
    """
    for arg in args:
        value = os.environ.get(arg)
        if value:
            return value
    return kwargs.get('default', '')


class OpenStackHelpFormatter(argparse.HelpFormatter):

    def __init__(self, prog, indent_increment=2, max_help_position=32,
                 width=None):
        super(OpenStackHelpFormatter, self).__init__(
            prog,
            indent_increment,
            max_help_position,
            width
        )

    def start_section(self, heading):
        # Title-case the headings.
        heading = '%s%s' % (heading[0].upper(), heading[1:])
        super(OpenStackHelpFormatter, self).start_section(heading)


class HelpAction(argparse.Action):
    """Custom help action.

    Provide a custom action so the -h and --help options
    to the main app will print a list of the commands.

    The commands are determined by checking the CommandManager
    instance, passed in as the "default" value for the action.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        outputs = []
        max_len = 0
        app = self.default
        parser.print_help(app.stdout)
        app.stdout.write('\nCommands for API v2 :\n')

        for name, ep in sorted(app.command_manager):
            factory = ep.load()
            cmd = factory(self, None)
            one_liner = cmd.get_description().split('\n')[0]
            outputs.append((name, one_liner))
            max_len = max(len(name), max_len)

        for (name, one_liner) in outputs:
            app.stdout.write('  %s  %s\n' % (name.ljust(max_len), one_liner))

        sys.exit(0)


class BashCompletionCommand(command.Command):
    """Prints all of the commands and options for bash-completion."""

    def take_action(self, parsed_args):
        commands = set()
        options = set()

        for option, _action in self.app.parser._option_string_actions.items():
            options.add(option)

        for command_name, _cmd in self.app.command_manager:
            commands.add(command_name)

        print(' '.join(commands | options))


class MistralShell(app.App):

    def __init__(self):
        super(MistralShell, self).__init__(
            description=__doc__.strip(),
            version=mistralclient.__version__,
            command_manager=commandmanager.CommandManager('mistral.cli'),
        )

        # Set v2 commands by default
        self._set_shell_commands(self._get_commands_v2())

    def configure_logging(self):
        log_lvl = logging.DEBUG if self.options.debug else logging.WARNING
        logging.basicConfig(
            format="%(levelname)s (%(module)s) %(message)s",
            level=log_lvl
        )
        logging.getLogger('iso8601').setLevel(logging.WARNING)

        if self.options.verbose_level <= 1:
            logging.getLogger('requests').setLevel(logging.WARNING)

    def build_option_parser(self, description, version,
                            argparse_kwargs=None):
        """Return an argparse option parser for this application.

        Subclasses may override this method to extend
        the parser with more global options.

        :param description: full description of the application
        :paramtype description: str
        :param version: version number for the application
        :paramtype version: str
        :param argparse_kwargs: extra keyword argument passed to the
                                ArgumentParser constructor
        :paramtype extra_kwargs: dict
        """
        argparse_kwargs = argparse_kwargs or {}

        parser = argparse.ArgumentParser(
            description=description,
            add_help=False,
            formatter_class=OpenStackHelpFormatter,
            **argparse_kwargs
        )

        parser.add_argument(
            '--version',
            action='version',
            version='%(prog)s {0}'.format(version),
            help='Show program\'s version number and exit.'
        )

        parser.add_argument(
            '-v', '--verbose',
            action='count',
            dest='verbose_level',
            default=self.DEFAULT_VERBOSE_LEVEL,
            help='Increase verbosity of output. Can be repeated.',
        )

        parser.add_argument(
            '--log-file',
            action='store',
            default=None,
            help='Specify a file to log output. Disabled by default.',
        )

        parser.add_argument(
            '-q', '--quiet',
            action='store_const',
            dest='verbose_level',
            const=0,
            help='Suppress output except warnings and errors.',
        )

        parser.add_argument(
            '-h', '--help',
            action=HelpAction,
            nargs=0,
            default=self,  # tricky
            help="Show this help message and exit.",
        )

        parser.add_argument(
            '--debug',
            default=False,
            action='store_true',
            help='Show tracebacks on errors.',
        )

        parser.add_argument(
            '--os-mistral-url',
            action='store',
            dest='mistral_url',
            default=env('OS_MISTRAL_URL'),
            help='Mistral API host (Env: OS_MISTRAL_URL)'
        )

        parser.add_argument(
            '--os-mistral-version',
            action='store',
            dest='mistral_version',
            default=env('OS_MISTRAL_VERSION', default='v2'),
            help='Mistral API version (default = v2) (Env: '
                 'OS_MISTRAL_VERSION)'
        )

        parser.add_argument(
            '--os-mistral-service-type',
            action='store',
            dest='service_type',
            default=env('OS_MISTRAL_SERVICE_TYPE', default='workflowv2'),
            help='Mistral service-type (should be the same name as in '
                 'keystone-endpoint) (default = workflowv2) (Env: '
                 'OS_MISTRAL_SERVICE_TYPE)'
        )

        parser.add_argument(
            '--os-mistral-endpoint-type',
            action='store',
            dest='endpoint_type',
            default=env('OS_MISTRAL_ENDPOINT_TYPE', default='publicURL'),
            help='Mistral endpoint-type (should be the same name as in '
                 'keystone-endpoint) (default = publicURL) (Env: '
                 'OS_MISTRAL_ENDPOINT_TYPE)'
        )

        parser.add_argument(
            '--os-username',
            action='store',
            dest='username',
            default=env('OS_USERNAME'),
            help='Authentication username (Env: OS_USERNAME)'
        )

        parser.add_argument(
            '--os-password',
            action='store',
            dest='password',
            default=env('OS_PASSWORD'),
            help='Authentication password (Env: OS_PASSWORD)'
        )

        parser.add_argument(
            '--os-tenant-id',
            action='store',
            dest='tenant_id',
            default=env('OS_TENANT_ID', 'OS_PROJECT_ID'),
            help='Authentication tenant identifier (Env: OS_TENANT_ID'
                 ' or OS_PROJECT_ID)'
        )

        parser.add_argument(
            '--os-project-id',
            action='store',
            dest='project_id',
            default=env('OS_TENANT_ID', 'OS_PROJECT_ID'),
            help='Authentication project identifier (Env: OS_TENANT_ID'
                 ' or OS_PROJECT_ID), will use tenant_id if both tenant_id'
                 ' and project_id are set'
        )

        parser.add_argument(
            '--os-tenant-name',
            action='store',
            dest='tenant_name',
            default=env('OS_TENANT_NAME', 'OS_PROJECT_NAME'),
            help='Authentication tenant name (Env: OS_TENANT_NAME'
                 ' or OS_PROJECT_NAME)'
        )

        parser.add_argument(
            '--os-project-name',
            action='store',
            dest='project_name',
            default=env('OS_TENANT_NAME', 'OS_PROJECT_NAME'),
            help='Authentication project name (Env: OS_TENANT_NAME'
                 ' or OS_PROJECT_NAME), will use tenant_name if both'
                 ' tenant_name and project_name are set'
        )

        parser.add_argument(
            '--os-auth-token',
            action='store',
            dest='token',
            default=env('OS_AUTH_TOKEN'),
            help='Authentication token (Env: OS_AUTH_TOKEN)'
        )

        parser.add_argument(
            '--os-project-domain-name',
            action='store',
            dest='project_domain_name',
            default=env('OS_PROJECT_DOMAIN_NAME'),
            help='Authentication project domain name or ID'
                 ' (Env: OS_PROJECT_DOMAIN_NAME or OS_PROJECT_DOMAIN_NAME)'
        )

        parser.add_argument(
            '--os-project-domain-id',
            action='store',
            dest='project_domain_id',
            default=env('OS_PROJECT_DOMAIN_ID'),
            help='Authentication project domain ID'
                 ' (Env: OS_PROJECT_DOMAIN_ID)'
        )

        parser.add_argument(
            '--os-user-domain-name',
            action='store',
            dest='user_domain_name',
            default=env('OS_USER_DOMAIN_NAME'),
            help='Authentication user domain name'
                 ' (Env: OS_USER_DOMAIN_NAME)'
        )

        parser.add_argument(
            '--os-user-domain-id',
            action='store',
            dest='user_domain_id',
            default=env('OS_USER_DOMAIN_ID'),
            help='Authentication user domain name'
                 ' (Env: OS_USER_DOMAIN_ID)'
        )

        parser.add_argument(
            '--os-auth-url',
            action='store',
            dest='auth_url',
            default=env('OS_AUTH_URL'),
            help='Authentication URL (Env: OS_AUTH_URL)'
        )

        parser.add_argument(
            '--os-cert',
            action='store',
            dest='os_cert',
            default=env('OS_CERT'),
            help='Client Certificate (Env: OS_CERT)'
        )

        parser.add_argument(
            '--os-key',
            action='store',
            dest='os_key',
            default=env('OS_KEY'),
            help='Client Key (Env: OS_KEY)'
        )

        parser.add_argument(
            '--os-cacert',
            action='store',
            dest='os_cacert',
            default=env('OS_CACERT'),
            help='Authentication CA Certificate (Env: OS_CACERT)'
        )

        parser.add_argument(
            '--os-region-name',
            action='store',
            dest='region_name',
            default=env('OS_REGION_NAME'),
            help='Region name (Env: OS_REGION_NAME)'
        )

        parser.add_argument(
            '--insecure',
            action='store_true',
            dest='insecure',
            default=env('MISTRALCLIENT_INSECURE', default=False),
            help='Disables SSL/TLS certificate verification '
                 '(Env: MISTRALCLIENT_INSECURE)'
        )

        parser.add_argument(
            '--auth-type',
            action='store',
            dest='auth_type',
            default=env('MISTRAL_AUTH_TYPE', default='keystone'),
            help='Authentication type. Valid options are: %s.'
                 ' (Env: MISTRAL_AUTH_TYPE)' % ', '.join(auth_types.ALL)
        )

        parser.add_argument(
            '--openid-client-id',
            action='store',
            dest='client_id',
            default=env('OPENID_CLIENT_ID'),
            help='Client ID (according to OpenID Connect).'
                 ' (Env: OPENID_CLIENT_ID)'
        )

        parser.add_argument(
            '--openid-client-secret',
            action='store',
            dest='client_secret',
            default=env('OPENID_CLIENT_SECRET'),
            help='Client secret (according to OpenID Connect)'
                 ' (Env: OPENID_CLIENT_SECRET)'
        )

        parser.add_argument(
            '--os-target-username',
            action='store',
            dest='target_username',
            default=env('OS_TARGET_USERNAME', default='admin'),
            help='Authentication username for target cloud'
                 ' (Env: OS_TARGET_USERNAME)'
        )

        parser.add_argument(
            '--os-target-password',
            action='store',
            dest='target_password',
            default=env('OS_TARGET_PASSWORD'),
            help='Authentication password for target cloud'
                 ' (Env: OS_TARGET_PASSWORD)'
        )

        parser.add_argument(
            '--os-target-tenant-id',
            action='store',
            dest='target_tenant_id',
            default=env('OS_TARGET_TENANT_ID'),
            help='Authentication tenant identifier for target cloud'
                 ' (Env: OS_TARGET_TENANT_ID)'
        )

        parser.add_argument(
            '--os-target-tenant-name',
            action='store',
            dest='target_tenant_name',
            default=env('OS_TARGET_TENANT_NAME'),
            help='Authentication tenant name for target cloud'
                 ' (Env: OS_TARGET_TENANT_NAME)'
        )

        parser.add_argument(
            '--os-target-auth-token',
            action='store',
            dest='target_token',
            default=env('OS_TARGET_AUTH_TOKEN'),
            help='Authentication token for target cloud'
                 ' (Env: OS_TARGET_AUTH_TOKEN)'
        )

        parser.add_argument(
            '--os-target-auth-url',
            action='store',
            dest='target_auth_url',
            default=env('OS_TARGET_AUTH_URL'),
            help='Authentication URL for target cloud'
                 ' (Env: OS_TARGET_AUTH_URL)'
        )

        parser.add_argument(
            '--os-target_cacert',
            action='store',
            dest='target_cacert',
            default=env('OS_TARGET_CACERT'),
            help='Authentication CA Certificate for target cloud'
                 ' (Env: OS_TARGET_CACERT)'
        )

        parser.add_argument(
            '--os-target-region-name',
            action='store',
            dest='target_region_name',
            default=env('OS_TARGET_REGION_NAME'),
            help='Region name for target cloud'
                 '(Env: OS_TARGET_REGION_NAME)'
        )

        parser.add_argument(
            '--os-target-user-domain-name',
            action='store',
            dest='target_user_domain_name',
            default=env('OS_TARGET_USER_DOMAIN_NAME'),
            help='User domain name for target cloud'
                 '(Env: OS_TARGET_USER_DOMAIN_NAME)'
        )

        parser.add_argument(
            '--os-target-user-domain-id',
            action='store',
            dest='target_user_domain_id',
            default=env('OS_TARGET_USER_DOMAIN_ID'),
            help='User domain ID for target cloud'
                 '(Env: OS_TARGET_USER_DOMAIN_ID)'
        )

        parser.add_argument(
            '--os-target-project-domain-name',
            action='store',
            dest='target_project_domain_name',
            default=env('OS_TARGET_PROJECT_DOMAIN_NAME'),
            help='Project domain name for target cloud'
                 '(Env: OS_TARGET_PROJECT_DOMAIN_NAME)'
        )

        parser.add_argument(
            '--os-target-project-domain-id',
            action='store',
            dest='target_project_domain_id',
            default=env('OS_TARGET_PROJECT_DOMAIN_ID'),
            help='Project domain ID for target cloud'
                 '(Env: OS_TARGET_PROJECT_DOMAIN_ID)'
        )

        parser.add_argument(
            '--target_insecure',
            action='store_true',
            dest='target_insecure',
            default=env('TARGET_MISTRALCLIENT_INSECURE', default=False),
            help='Disables SSL/TLS certificate verification for target cloud '
                 '(Env: TARGET_MISTRALCLIENT_INSECURE)'
        )

        parser.add_argument(
            '--profile',
            dest='profile',
            metavar='HMAC_KEY',
            default=env('OS_PROFILE'),
            help='HMAC key to use for encrypting context data for performance '
                 'profiling of operation. This key should be one of the '
                 'values configured for the osprofiler middleware in mistral, '
                 'it is specified in the profiler section of the mistral '
                 'configuration (i.e. /etc/mistral/mistral.conf). Without the '
                 'key, profiling will not be triggered even if osprofiler is '
                 'enabled on the server side.'
        )

        return parser

    def initialize_app(self, argv):
        self._clear_shell_commands()

        ver = client.determine_client_version(self.options.mistral_version)

        self._set_shell_commands(self._get_commands(ver))

        # bash-completion and help messages should not require client creation
        need_client = not (
            ('bash-completion' in argv) or
            ('help' in argv) or
            ('-h' in argv) or
            ('--help' in argv) or
            not argv)

        # Set default for auth_url if not supplied. The default is not
        # set at the parser to support use cases where auth is not enabled.
        # An example use case would be a developer's environment.
        if not self.options.auth_url:
            if self.options.password or self.options.token:
                self.options.auth_url = 'http://localhost:35357/v3'

        if (self.options.auth_type == 'keystone' and
                not self.options.auth_url.endswith("/v2.0")):
            # Assume that keystone V3 is used and try to be more user-friendly,
            # i.e provide default values for domains
            if (not self.options.project_domain_id and
                    not self.options.project_domain_name):
                self.options.project_domain_id = "default"
            if (not self.options.user_domain_id and
                    not self.options.user_domain_name):
                self.options.user_domain_id = "default"
            if (not self.options.target_project_domain_id and
                    not self.options.target_project_domain_name):
                self.options.target_project_domain_id = "default"
            if (not self.options.target_user_domain_id and
                    not self.options.target_user_domain_name):
                self.options.target_user_domain_id = "default"

        if self.options.auth_url and not self.options.token:
            if not self.options.username:
                raise exe.IllegalArgumentException(
                    ("You must provide a username "
                     "via --os-username env[OS_USERNAME]")
                )

            if not self.options.password:
                raise exe.IllegalArgumentException(
                    ("You must provide a password "
                     "via --os-password env[OS_PASSWORD]")
                )
        self.client = self._create_client() if need_client else None

        # Adding client_manager variable to make mistral client work with
        # unified OpenStack client.
        ClientManager = type(
            'ClientManager',
            (object,),
            dict(workflow_engine=self.client)
        )

        self.client_manager = ClientManager()

    def _create_client(self):
        kwargs = {
            'cert': self.options.os_cert,
            'key': self.options.os_key,
            'user_domain_name': self.options.user_domain_name,
            'user_domain_id': self.options.user_domain_id,
            'project_domain_name': self.options.project_domain_name,
            'project_domain_id': self.options.project_domain_id,
            'target_project_domain_name':
                self.options.target_project_domain_name,
            'target_project_domain_id': self.options.target_project_domain_id,
            'target_user_domain_name': self.options.target_user_domain_name,
            'target_user_domain_id': self.options.target_user_domain_id
        }

        return client.client(
            mistral_url=self.options.mistral_url,
            username=self.options.username,
            api_key=self.options.password,
            project_name=self.options.tenant_name or self.options.project_name,
            auth_url=self.options.auth_url,
            project_id=self.options.tenant_id or self.options.project_id,
            endpoint_type=self.options.endpoint_type,
            service_type=self.options.service_type,
            region_name=self.options.region_name,
            auth_token=self.options.token,
            cacert=self.options.os_cacert,
            insecure=self.options.insecure,
            profile=self.options.profile,
            auth_type=self.options.auth_type,
            client_id=self.options.client_id,
            client_secret=self.options.client_secret,
            target_username=self.options.target_username,
            target_api_key=self.options.target_password,
            target_project_name=self.options.target_tenant_name,
            target_auth_url=self.options.target_auth_url,
            target_project_id=self.options.target_tenant_id,
            target_auth_token=self.options.target_token,
            target_cacert=self.options.target_cacert,
            target_region_name=self.options.target_region_name,
            target_insecure=self.options.target_insecure,
            **kwargs
        )

    def _set_shell_commands(self, cmds_dict):
        for k, v in cmds_dict.items():
            self.command_manager.add_command(k, v)

    def _clear_shell_commands(self):
        exclude_cmds = ['help', 'complete']

        cmds = self.command_manager.commands.copy()
        for k, v in cmds.items():
            if k not in exclude_cmds:
                self.command_manager.commands.pop(k)

    def _get_commands(self, version):
        if version == 2:
            return self._get_commands_v2()

        return {}

    @staticmethod
    def _get_commands_v2():
        return {
            'bash-completion': BashCompletionCommand,
            'workbook-list': mistralclient.commands.v2.workbooks.List,
            'workbook-get': mistralclient.commands.v2.workbooks.Get,
            'workbook-create': mistralclient.commands.v2.workbooks.Create,
            'workbook-delete': mistralclient.commands.v2.workbooks.Delete,
            'workbook-update': mistralclient.commands.v2.workbooks.Update,
            'workbook-get-definition':
            mistralclient.commands.v2.workbooks.GetDefinition,
            'workbook-validate': mistralclient.commands.v2.workbooks.Validate,
            'workflow-list': mistralclient.commands.v2.workflows.List,
            'workflow-get': mistralclient.commands.v2.workflows.Get,
            'workflow-create': mistralclient.commands.v2.workflows.Create,
            'workflow-delete': mistralclient.commands.v2.workflows.Delete,
            'workflow-update': mistralclient.commands.v2.workflows.Update,
            'workflow-get-definition':
            mistralclient.commands.v2.workflows.GetDefinition,
            'workflow-validate': mistralclient.commands.v2.workflows.Validate,
            'environment-create':
            mistralclient.commands.v2.environments.Create,
            'environment-delete':
            mistralclient.commands.v2.environments.Delete,
            'environment-update':
            mistralclient.commands.v2.environments.Update,
            'environment-list': mistralclient.commands.v2.environments.List,
            'environment-get': mistralclient.commands.v2.environments.Get,
            'run-action': mistralclient.commands.v2.action_executions.Create,
            'action-execution-list':
            mistralclient.commands.v2.action_executions.List,
            'action-execution-get':
            mistralclient.commands.v2.action_executions.Get,
            'action-execution-get-input':
            mistralclient.commands.v2.action_executions.GetInput,
            'action-execution-get-output':
            mistralclient.commands.v2.action_executions.GetOutput,
            'action-execution-update':
            mistralclient.commands.v2.action_executions.Update,
            'action-execution-delete':
            mistralclient.commands.v2.action_executions.Delete,
            'execution-create': mistralclient.commands.v2.executions.Create,
            'execution-delete': mistralclient.commands.v2.executions.Delete,
            'execution-update': mistralclient.commands.v2.executions.Update,
            'execution-list': mistralclient.commands.v2.executions.List,
            'execution-get': mistralclient.commands.v2.executions.Get,
            'execution-get-input':
            mistralclient.commands.v2.executions.GetInput,
            'execution-get-output':
            mistralclient.commands.v2.executions.GetOutput,
            'task-list': mistralclient.commands.v2.tasks.List,
            'task-get': mistralclient.commands.v2.tasks.Get,
            'task-get-published': mistralclient.commands.v2.tasks.GetPublished,
            'task-get-result': mistralclient.commands.v2.tasks.GetResult,
            'task-rerun': mistralclient.commands.v2.tasks.Rerun,
            'action-list': mistralclient.commands.v2.actions.List,
            'action-get': mistralclient.commands.v2.actions.Get,
            'action-create': mistralclient.commands.v2.actions.Create,
            'action-delete': mistralclient.commands.v2.actions.Delete,
            'action-update': mistralclient.commands.v2.actions.Update,
            'action-get-definition':
            mistralclient.commands.v2.actions.GetDefinition,
            'action-validate': mistralclient.commands.v2.actions.Validate,
            'cron-trigger-list': mistralclient.commands.v2.cron_triggers.List,
            'cron-trigger-get': mistralclient.commands.v2.cron_triggers.Get,
            'cron-trigger-create':
            mistralclient.commands.v2.cron_triggers.Create,
            'cron-trigger-delete':
            mistralclient.commands.v2.cron_triggers.Delete,
            'event-trigger-list':
            mistralclient.commands.v2.event_triggers.List,
            'event-trigger-get': mistralclient.commands.v2.event_triggers.Get,
            'event-trigger-create':
            mistralclient.commands.v2.event_triggers.Create,
            'event-trigger-delete':
            mistralclient.commands.v2.event_triggers.Delete,
            'service-list': mistralclient.commands.v2.services.List,
            'member-create': mistralclient.commands.v2.members.Create,
            'member-delete': mistralclient.commands.v2.members.Delete,
            'member-update': mistralclient.commands.v2.members.Update,
            'member-list': mistralclient.commands.v2.members.List,
            'member-get': mistralclient.commands.v2.members.Get,
        }


def main(argv=sys.argv[1:]):
    return MistralShell().run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



# Copyright 2016 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import pprint

import requests
from six.moves import urllib

from mistralclient import auth


LOG = logging.getLogger(__name__)


class KeycloakAuthHandler(auth.AuthHandler):

    def authenticate(self, req, session=None):
        """Performs authentication using Keycloak OpenID Protocol.

        :param req: Request dict containing list of parameters required
            for Keycloak authentication.

            * auth_url: Base authentication url of KeyCloak server (e.g.
                "https://my.keycloak:8443/auth"
            * client_id: Client ID (according to OpenID Connect protocol).
            * client_secret: Client secret (according to OpenID Connect
                protocol).
            * project_name: KeyCloak realm name.
            * username: User name (Optional, if None then access_token must be
                provided).
            * api_key: Password (Optional).
            * access_token: Access token. If passed, username and password are
                not used and this method just validates the token and refreshes
                it if needed (Optional, if None then username must be
                provided).
            * cacert: SSL certificate file (Optional).
            * insecure: If True, SSL certificate is not verified (Optional).

        :param session: Keystone session object. Not used by this plugin.

        """

        if not isinstance(req, dict):
            raise TypeError('The input "req" is not typeof dict.')

        auth_url = req.get('auth_url')
        client_id = req.get('client_id')
        client_secret = req.get('client_secret')
        realm_name = req.get('project_name')
        username = req.get('username')
        password = req.get('api_key')
        access_token = req.get('access_token')
        cacert = req.get('cacert')
        insecure = req.get('insecure', False)

        if not auth_url:
            raise ValueError('Base authentication url is not provided.')

        if not client_id:
            raise ValueError('Client ID is not provided.')

        if not realm_name:
            raise ValueError('Project(realm) name is not provided.')

        if username and access_token:
            raise ValueError(
                "User name and access token can't be "
                "provided at the same time."
            )

        if not username and not access_token:
            raise ValueError(
                'Either user name or access token must be provided.'
            )

        if access_token:
            response = self._authenticate_with_token(
                auth_url,
                client_id,
                client_secret,
                access_token,
                cacert,
                insecure
            )
        else:
            response = self._authenticate_with_password(
                auth_url,
                client_id,
                client_secret,
                realm_name,
                username,
                password,
                cacert,
                insecure
            )

        return {'auth_token': response, 'project_id': realm_name}

    @staticmethod
    def _authenticate_with_token(auth_url, client_id, client_secret,
                                 auth_token, cacert=None, insecure=None):
        # TODO(rakhmerov): Implement.
        raise NotImplementedError

    @staticmethod
    def _authenticate_with_password(auth_url, client_id, client_secret,
                                    realm_name, username, password,
                                    cacert=None, insecure=None):
        access_token_endpoint = (
            "%s/realms/%s/protocol/openid-connect/token" %
            (auth_url, realm_name)
        )

        verify = None
        if urllib.parse.urlparse(access_token_endpoint).scheme == "https":
            verify = False if insecure else cacert if cacert else True

        body = {
            'grant_type': 'password',
            'username': username,
            'password': password,
            'client_id': client_id,
            'scope': 'profile'
        }

        if client_secret:
            body['client_secret'] = client_secret,

        resp = requests.post(
            access_token_endpoint,
            data=body,
            verify=verify
        )

        try:
            resp.raise_for_status()
        except Exception as e:
            raise Exception("Failed to get access token:\n %s" % str(e))

        LOG.debug("HTTP response from OIDC provider: %s",
                  pprint.pformat(resp.json()))

        return resp.json()['access_token']


def get_system_ca_file():
    """Return path to system default CA file."""
    # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
    # Suse, FreeBSD/OpenBSD, MacOSX, and the bundled ca
    ca_path = ['/etc/ssl/certs/ca-certificates.crt',
               '/etc/pki/tls/certs/ca-bundle.crt',
               '/etc/ssl/ca-bundle.pem',
               '/etc/ssl/cert.pem',
               '/System/Library/OpenSSL/certs/cacert.pem',
               requests.certs.where()]
    for ca in ca_path:
        LOG.debug("Looking for ca file %s", ca)
        if os.path.exists(ca):
            LOG.debug("Using ca file %s", ca)
            return ca
    LOG.warning("System ca file could not be found.")

# An example of working curl request to keycloak
# curl -d "client_id=admin-cli" -d "client_secret=secret"
# -d "username=admin" -d "password=qwerty" -d "grant_type=password"
# "http://localhost:8080/auth/realms/master/protocol/openid-connect/token"

# An example of using KeyCloak OpenID authentication.
if __name__ == '__main__':
    print("Using username/password to get access token from KeyCloak...")

    auth_handler = KeycloakAuthHandler()

    a_token = auth_handler.authenticate(
        dict(
            "https://my.keycloak:8443/auth",
            client_id="mistral_client",
            client_secret="secret",
            project_name="mistral",
            username="user",
            api_key="secret",
            insecure=True
        )
    )['auth_token']

    print("Auth token: %s" % a_token)



# Copyright 2016 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

import keystoneauth1.access.service_catalog as sc
import keystoneauth1.identity.generic as auth_plugin
from keystoneauth1 import session as ks_session
import mistralclient.api.httpclient as api
from mistralclient import auth as mistral_auth
from oslo_serialization import jsonutils


LOG = logging.getLogger(__name__)


class KeystoneAuthHandler(mistral_auth.AuthHandler):

    def authenticate(self, req, session=None):
        """Perform authentication via Keystone.

        :param req: Request dict containing the parameters required
            for Keystone authentication.
        """
        reqs = self._separate_target_reqs(req)
        try:
            result = self._authenticate(reqs, session)
        except Exception as e:
            if "Cannot use v2 authentication with domain scope" in str(e):
                LOG.warning("Client tried to use v2 authentication with "
                            "domain scope. Domain parameters are assumed "
                            "to be erroneously set. Retrying "
                            "authentication without them. "
                            "Request parameters: %s" % str(reqs))
                domainless_reqs = [reqs[0],
                                   self._remove_domain(reqs[1])]
                result = self._authenticate(domainless_reqs, session)
            else:
                raise
        return result

    @staticmethod
    def _separate_target_reqs(req):
        """Separates parameters into target and non-target ones.

        target_* parameters are rekeyed by removing the prefix.

        :param req: Request dict containing the parameters for Keystone
        authentication.
        :return: list of [non-target, target] request parameters
        """
        r = {}
        target_r = {}
        target_prefix = "target_"
        for key in req:
            if key.startswith(target_prefix):
                target_r[key[len(target_prefix):]] = req[key]
            else:
                r[key] = req[key]

        return [r, target_r]

    @staticmethod
    def _remove_domain(req):
        """Remove all domain parameters from req.

        Keystoneauth with V2 does not accept domain parameters. This
        is an incompatible change from Keystoneclient but it would
        unnecessarily break clients of Mistral. It is safe to remove
        domain parameters if V2 auth is targeted.

        :param req: Request dict containing list of parameters required
        :return: Request dict without domains
        """
        r = {}
        for key in req:
            if "domain" not in key:
                r[key] = req[key]
        return r

    @staticmethod
    def _get_auth(api_key=None, auth_token=None, auth_url=None,
                  project_domain_id=None, project_domain_name=None,
                  project_id=None, project_name=None, user_domain_id=None,
                  user_domain_name=None, user_id=None, username=None,
                  **kwargs):

        if project_name and project_id:
            raise RuntimeError(
                'Only one of project_name or project_id should be set'
            )

        if username and user_id:
            raise RuntimeError(
                'Only one of username or user_id should be set'
            )

        auth = {}

        if auth_token:
            auth = auth_plugin.Token(
                auth_url=auth_url,
                project_domain_id=project_domain_id,
                project_domain_name=project_domain_name,
                project_id=project_id,
                project_name=project_name,
                token=auth_token
            )
        elif api_key and (username or user_id):
            auth = auth_plugin.Password(
                auth_url=auth_url,
                password=api_key,
                project_domain_id=project_domain_id,
                project_domain_name=project_domain_name,
                project_id=project_id,
                project_name=project_name,
                user_domain_id=user_domain_id,
                user_domain_name=user_domain_name,
                user_id=user_id,
                username=username
            )

        return auth

    def _authenticate(self, reqs, session=None):
        """Performs authentication via Keystone.

        :param reqs: Request dict containing list of parameters required
            for Keystone authentication.
        :return: Auth response dict
        """
        if not isinstance(reqs[0], dict):
            raise TypeError('The input "req" is not typeof dict.')
        if not isinstance(reqs[1], dict):
            raise TypeError('The input "req" is not typeof dict.')

        auth_response = {}
        req = reqs[0]
        cacert = req.get('cacert')
        endpoint_type = req.get('endpoint_type', 'publicURL')
        insecure = req.get('insecure')
        mistral_url = req.get('mistral_url')
        region_name = req.get('region_name')
        service_type = req.get('service_type', 'workflowv2')

        verify = self._verification_needed(cacert, insecure)

        if not session:
            auth = self._get_auth(**req)

            if auth:
                session = ks_session.Session(auth=auth, verify=verify)

        if session:
            if not mistral_url:
                try:
                    mistral_url = session.get_endpoint(
                        service_type=service_type,
                        endpoint_type=endpoint_type,
                        region_name=region_name
                    )
                except Exception:
                    mistral_url = None

            auth_response['mistral_url'] = mistral_url
            auth_response['session'] = session

        target_req = reqs[1]

        if "auth_url" in target_req:
            target_auth = self._get_auth(**target_req)

            if target_auth:

                # target cacert and insecure
                cacert = target_req.get('cacert')
                insecure = target_req.get('insecure')

                verify = self._verification_needed(cacert, insecure)

                target_session = ks_session.Session(
                    auth=target_auth,
                    verify=verify
                )

                target_auth_headers = target_session.get_auth_headers() or {}

                target_auth_token = target_auth_headers.get('X-Auth-Token')

                auth_response.update({
                    api.TARGET_AUTH_TOKEN: target_auth_token,
                    api.TARGET_PROJECT_ID: target_session.get_project_id(),
                    api.TARGET_USER_ID: target_session.get_user_id(),
                    api.TARGET_AUTH_URI: target_auth._plugin.auth_url,
                })

                access = target_auth.get_access(target_session)
                service_catalog = access.service_catalog

                if self._is_service_catalog_v2(service_catalog):
                    access_data = access._data["access"]
                    if not len(access_data['serviceCatalog']):
                        LOG.warning(
                            "Service Catalog empty, some authentication"
                            "credentials may be missing. This can cause"
                            "malfunction in the Mistral action executions.")
                    sc_json = jsonutils.dumps(access_data)
                    auth_response[api.TARGET_SERVICE_CATALOG] = sc_json

        if not auth_response:
            LOG.debug("No valid token or password + user provided. "
                      "Continuing without authentication")
            return {}

        return auth_response

    @staticmethod
    def _verification_needed(cacert, insecure):
        """Return the verify parameter.

        The value of verify can be either True/False or a cacert.

        :param cacert None or path to CA cert file
        :param insecure truthy value to switch on SSL verification
        """
        if insecure is False or insecure is None:
            verify = cacert or True
        else:
            verify = False
        return verify

    @staticmethod
    def _is_service_catalog_v2(catalog):
        """Check if the service catalog is of type ServiceCatalogV2

        :param catalog: the service catalog
        :return: True if V2, False otherwise
        """
        return type(catalog) is sc.ServiceCatalogV2



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json

from mistralclient.api.v2 import tasks
from mistralclient.tests.unit.v2 import base

# TODO(everyone): later we need additional tests verifying all the errors etc.

TASK = {
    'id': "1",
    'workflow_execution_id': '123',
    'name': 'my_task',
    'workflow_name': 'my_wf',
    'state': 'RUNNING',
    'tags': ['deployment', 'demo'],
    'result': {'some': 'result'}
}


URL_TEMPLATE = '/tasks'
URL_TEMPLATE_ID = '/tasks/%s'


class TestTasksV2(base.BaseClientV2Test):
    def test_list(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE,
                               json={'tasks': [TASK]})

        task_list = self.tasks.list()

        self.assertEqual(1, len(task_list))
        task = task_list[0]

        self.assertEqual(
            tasks.Task(self.tasks, TASK).to_dict(),
            task.to_dict()
        )

    def test_list_with_fields(self):
        field_params = "?fields=id,name"

        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE + field_params,
                               json={'tasks': [TASK]})

        self.tasks.list(fields=["id,name"])
        self.assertTrue(self.requests_mock.called_once)

    def test_list_with_no_limit(self):
        self.requests_mock.get(self.TEST_URL + URL_TEMPLATE,
                               json={'tasks': [TASK]})

        task_list = self.tasks.list(limit=-1)

        self.assertEqual(1, len(task_list))

        last_request = self.requests_mock.last_request

        self.assertNotIn('limit', last_request.qs)

    def test_get(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % TASK['id']
        self.requests_mock.get(url, json=TASK)

        task = self.tasks.get(TASK['id'])

        self.assertEqual(
            tasks.Task(self.tasks, TASK).to_dict(),
            task.to_dict()
        )

    def test_rerun(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % TASK['id']
        self.requests_mock.put(url, json=TASK)

        task = self.tasks.rerun(TASK['id'])

        self.assertDictEqual(
            tasks.Task(self.tasks, TASK).to_dict(),
            task.to_dict()
        )

        body = {
            'reset': True,
            'state': 'RUNNING',
            'id': TASK['id']
        }
        self.assertDictEqual(body, self.requests_mock.last_request.json())

    def test_rerun_no_reset(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % TASK['id']
        self.requests_mock.put(url, json=TASK)

        task = self.tasks.rerun(TASK['id'], reset=False)

        self.assertDictEqual(
            tasks.Task(self.tasks, TASK).to_dict(),
            task.to_dict()
        )

        body = {
            'reset': False,
            'state': 'RUNNING',
            'id': TASK['id']
        }
        self.assertDictEqual(body, self.requests_mock.last_request.json())

    def test_rerun_update_env(self):
        url = self.TEST_URL + URL_TEMPLATE_ID % TASK['id']
        self.requests_mock.put(url, json=TASK)

        task = self.tasks.rerun(TASK['id'], env={'k1': 'foobar'})

        self.assertDictEqual(
            tasks.Task(self.tasks, TASK).to_dict(),
            task.to_dict()
        )

        body = {
            'reset': True,
            'state': 'RUNNING',
            'id': TASK['id'],
            'env': json.dumps({'k1': 'foobar'})
        }
        self.assertDictEqual(body, self.requests_mock.last_request.json())



# Copyright 2015 - Huawei Technologies Co. Ltd
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import os
import yaml

from six.moves.urllib import parse
from six.moves.urllib import request

from mistralclient import exceptions


def do_action_on_many(action, resources, success_msg, error_msg):
    """Helper to run an action on many resources."""
    failure_flag = False

    for resource in resources:
        try:
            action(resource)
            print(success_msg % resource)
        except Exception as e:
            failure_flag = True
            print(e)

    if failure_flag:
        raise exceptions.MistralClientException(error_msg)


def load_content(content):
    if content is None or content == '':
        return dict()

    try:
        data = yaml.safe_load(content)
    except Exception:
        data = json.loads(content)

    return data


def load_file(path):
    with open(path, 'r') as f:
        return load_content(f.read())


def get_contents_if_file(contents_or_file_name):
    """Get the contents of a file.

    If the value passed in is a file name or file URI, return the
    contents. If not, or there is an error reading the file contents,
    return the value passed in as the contents.

    For example, a workflow definition will be returned if either the
    workflow definition file name, or file URI are passed in, or the
    actual workflow definition itself is passed in.
    """
    try:
        if parse.urlparse(contents_or_file_name).scheme:
            definition_url = contents_or_file_name
        else:
            path = os.path.abspath(contents_or_file_name)
            definition_url = parse.urljoin(
                'file:',
                request.pathname2url(path)
            )
        return request.urlopen(definition_url).read().decode('utf8')
    except Exception:
        return contents_or_file_name


def load_json(input_string):
    try:
        with open(input_string) as fh:
            return json.load(fh)
    except IOError:
        return json.loads(input_string)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json

import mock

from mistralclient.api.v2 import tasks
from mistralclient.commands.v2 import tasks as task_cmd
from mistralclient.tests.unit import base

TASK_DICT = {
    'id': '123',
    'name': 'some',
    'workflow_name': 'thing',
    'workflow_namespace': '',
    'workflow_execution_id': '321',
    'state': 'RUNNING',
    'state_info': None,
    'created_at': '1',
    'updated_at': '1',
}

TASK_RESULT = {"test": "is", "passed": "successfully"}
TASK_PUBLISHED = {"bar1": "val1", "var2": 2}

TASK_WITH_RESULT_DICT = TASK_DICT.copy()
TASK_WITH_RESULT_DICT.update({'result': json.dumps(TASK_RESULT)})
TASK_WITH_PUBLISHED_DICT = TASK_DICT.copy()
TASK_WITH_PUBLISHED_DICT.update({'published': json.dumps(TASK_PUBLISHED)})

TASK = tasks.Task(mock, TASK_DICT)
TASK_WITH_RESULT = tasks.Task(mock, TASK_WITH_RESULT_DICT)
TASK_WITH_PUBLISHED = tasks.Task(mock, TASK_WITH_PUBLISHED_DICT)

EXPECTED_TASK_RESULT = (
    '123', 'some', 'thing', '', '321', 'RUNNING', None, '1', '1'
)


class TestCLITasksV2(base.BaseCommandTest):
    def test_list(self):
        self.client.tasks.list.return_value = [TASK]

        result = self.call(task_cmd.List)

        self.assertEqual([EXPECTED_TASK_RESULT], result[1])
        self.assertEqual(
            self.client.tasks.list.call_args[1]["fields"],
            task_cmd.TaskFormatter.COLUMN_FIELD_NAMES
        )

    def test_list_with_workflow_execution(self):
        self.client.tasks.list.return_value = [TASK]

        result = self.call(task_cmd.List, app_args=['workflow_execution'])

        self.assertEqual([EXPECTED_TASK_RESULT], result[1])

    def test_get(self):
        self.client.tasks.get.return_value = TASK

        result = self.call(task_cmd.Get, app_args=['id'])

        self.assertEqual(EXPECTED_TASK_RESULT, result[1])

    def test_get_result(self):
        self.client.tasks.get.return_value = TASK_WITH_RESULT

        self.call(task_cmd.GetResult, app_args=['id'])

        self.assertDictEqual(
            TASK_RESULT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_get_published(self):
        self.client.tasks.get.return_value = TASK_WITH_PUBLISHED

        self.call(task_cmd.GetPublished, app_args=['id'])

        self.assertDictEqual(
            TASK_PUBLISHED,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_rerun(self):
        self.client.tasks.rerun.return_value = TASK

        result = self.call(task_cmd.Rerun, app_args=['id'])

        self.assertEqual(EXPECTED_TASK_RESULT, result[1])

    def test_rerun_no_reset(self):
        self.client.tasks.rerun.return_value = TASK

        result = self.call(task_cmd.Rerun, app_args=['id', '--resume'])

        self.assertEqual(EXPECTED_TASK_RESULT, result[1])

    def test_rerun_update_env(self):
        self.client.tasks.rerun.return_value = TASK

        result = self.call(
            task_cmd.Rerun,
            app_args=['id', '--env', '{"k1": "foobar"}']
        )

        self.assertEqual(EXPECTED_TASK_RESULT, result[1])

    def test_rerun_no_reset_update_env(self):
        self.client.tasks.rerun.return_value = TASK

        result = self.call(
            task_cmd.Rerun,
            app_args=['id', '--resume', '--env', '{"k1": "foobar"}']
        )

        self.assertEqual(EXPECTED_TASK_RESULT, result[1])



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mistralclient.api.v2 import client
from mistralclient.tests.unit import base


class BaseClientV2Test(base.BaseClientTest):

    TEST_URL = 'http://mistral.example.com'

    def setUp(self):
        super(BaseClientV2Test, self).setUp()

        with mock.patch(
                'mistralclient.auth.keystone.KeystoneAuthHandler.authenticate',
                return_value={'session': None}):
            self._client = client.Client(project_name="test",
                                         mistral_url=self.TEST_URL)
            self.workbooks = self._client.workbooks
            self.executions = self._client.executions
            self.tasks = self._client.tasks
            self.workflows = self._client.workflows
            self.environments = self._client.environments
            self.action_executions = self._client.action_executions
            self.actions = self._client.actions
            self.services = self._client.services
            self.members = self._client.members



# Copyright (c) 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import time

from tempest.lib import exceptions

from mistralclient.tests.functional.cli import base


MISTRAL_URL = "http://localhost:8989/v2"


class MistralClientTestBase(base.MistralCLIAuth, base.MistralCLIAltAuth):

    _mistral_url = MISTRAL_URL

    @classmethod
    def setUpClass(cls):
        super(MistralClientTestBase, cls).setUpClass()

        cls.wb_def = os.path.relpath(
            'functionaltests/resources/v2/wb_v2.yaml', os.getcwd()
        )

        cls.wb_with_tags_def = os.path.relpath(
            'functionaltests/resources/v2/wb_with_tags_v2.yaml', os.getcwd()
        )

        cls.wf_def = os.path.relpath(
            'functionaltests/resources/v2/wf_v2.yaml', os.getcwd()
        )

        cls.wf_single_def = os.path.relpath(
            'functionaltests/resources/v2/wf_single_v2.yaml', os.getcwd()
        )

        cls.wf_with_delay_def = os.path.relpath(
            'functionaltests/resources/v2/wf_delay_v2.yaml', os.getcwd()
        )

        cls.wf_wrapping_wf = os.path.relpath(
            'functionaltests/resources/v2/wf_wrapping_wf_v2.yaml', os.getcwd()
        )
        cls.top_level_wf = os.path.relpath(
            'functionaltests/resources/v2/for_namespaces/top_level_wf.yaml',
            os.getcwd()
        )
        cls.middle_wf = os.path.relpath(
            'functionaltests/resources/v2/for_namespaces/middle_wf.yaml',
            os.getcwd()
        )
        cls.lowest_level_wf = os.path.relpath(
            'functionaltests/resources/v2/for_namespaces/lowest_level_wf.yaml',
            os.getcwd()
        )
        cls.async_wf_def = os.path.relpath(
            'functionaltests/resources/v2/async.yaml',
            os.getcwd()
        )

        cls.act_def = os.path.relpath(
            'functionaltests/resources/v2/action_v2.yaml', os.getcwd()
        )

        cls.act_tag_def = os.path.relpath(
            'functionaltests/resources/v2/action_v2_tags.yaml', os.getcwd()
        )

    def setUp(self):
        super(MistralClientTestBase, self).setUp()

    def get_field_value(self, obj, field):
        return [
            o['Value'] for o in obj
            if o['Field'] == "{0}".format(field)
        ][0]

    def get_item_info(self, get_from, get_by, value):
        return [i for i in get_from if i[get_by] == value][0]

    def mistral_admin(self, cmd, params=""):
        self.clients = self._get_admin_clients()

        return self.parser.listing(
            self.mistral('{0}'.format(cmd), params='{0}'.format(params))
        )

    def mistral_alt_user(self, cmd, params=""):
        self.clients = self._get_alt_clients()

        return self.parser.listing(
            self.mistral_alt('{0}'.format(cmd), params='{0}'.format(params))
        )

    def mistral_cli(self, admin, cmd, params=''):
        if admin:
            return self.mistral_admin(cmd, params)
        else:
            return self.mistral_alt_user(cmd, params)

    def workbook_create(self, wb_def, admin=True, scope='private'):
        params = '{0}'.format(wb_def)

        if scope == 'public':
            params += ' --public'

        wb = self.mistral_cli(
            admin,
            'workbook-create',
            params=params
        )

        wb_name = self.get_field_value(wb, "Name")

        self.addCleanup(
            self.mistral_cli,
            admin,
            'workbook-delete',
            params=wb_name
        )

        self.addCleanup(
            self.mistral_cli,
            admin,
            'workflow-delete',
            params='wb.wf1'
        )

        return wb

    def workflow_create(self, wf_def, namespace='', admin=True,
                        scope='private'):
        params = '{0}'.format(wf_def)

        if scope == 'public':
            params += ' --public'

        if namespace:
            params += " --namespace " + namespace

        wf = self.mistral_cli(
            admin,
            'workflow-create',
            params=params
        )

        for workflow in wf:
            self.addCleanup(
                self.mistral_cli,
                admin,
                'workflow-delete',
                params=workflow['ID']
            )

        return wf

    def workflow_member_create(self, wf_id):
        cmd_param = (
            '%s workflow %s' % (wf_id, self.get_project_id("demo"))
        )
        member = self.mistral_admin("member-create", params=cmd_param)

        self.addCleanup(
            self.mistral_admin,
            'member-delete',
            params=cmd_param
        )

        return member

    def action_create(self, act_def, admin=True, scope='private'):
        params = '{0}'.format(act_def)

        if scope == 'public':
            params += ' --public'

        acts = self.mistral_cli(
            admin,
            'action-create',
            params=params
        )

        for action in acts:
            self.addCleanup(
                self.mistral_cli,
                admin,
                'action-delete',
                params=action['Name']
            )

        return acts

    def cron_trigger_create(self, name, wf_name, wf_input, pattern=None,
                            count=None, first_time=None, admin=True):
        optional_params = ""

        if pattern:
            optional_params += ' --pattern "{}"'.format(pattern)
        if count:
            optional_params += ' --count {}'.format(count)
        if first_time:
            optional_params += ' --first-time "{}"'.format(first_time)

        trigger = self.mistral_cli(
            admin,
            'cron-trigger-create',
            params='{} {} {} {}'.format(name, wf_name, wf_input,
                                        optional_params))

        self.addCleanup(self.mistral_cli,
                        admin,
                        'cron-trigger-delete',
                        params=name)

        return trigger

    def event_trigger_create(self, name, wf_id, exchange,
                             topic, event, wf_input, admin=True):

        trigger = self.mistral_cli(
            admin,
            'event-trigger-create',
            params=' '.join((name, wf_id, exchange, topic, event, wf_input)))
        ev_tr_id = self.get_field_value(trigger, 'ID')
        self.addCleanup(self.mistral_cli,
                        admin,
                        'event-trigger-delete',
                        params=ev_tr_id)

        return trigger

    def execution_create(self, params, admin=True):
        ex = self.mistral_cli(admin, 'execution-create', params=params)
        exec_id = self.get_field_value(ex, 'ID')

        self.addCleanup(
            self.mistral_cli,
            admin,
            'execution-delete',
            params="{} --force".format(exec_id)
        )

        return ex

    def environment_create(self, params, admin=True):
        env = self.mistral_cli(admin, 'environment-create', params=params)
        env_name = self.get_field_value(env, 'Name')

        self.addCleanup(
            self.mistral_cli,
            admin,
            'environment-delete',
            params=env_name
        )

        return env

    def create_file(self, file_name, file_body=""):
        f = open(file_name, 'w')
        f.write(file_body)
        f.close()

        self.addCleanup(os.remove, file_name)

    def wait_execution_success(self, exec_id, timeout=180):
        start_time = time.time()

        ex = self.mistral_admin('execution-get', params=exec_id)
        exec_state = self.get_field_value(ex, 'State')

        expected_states = ['SUCCESS', 'RUNNING']

        while exec_state != 'SUCCESS':
            if time.time() - start_time > timeout:
                msg = ("Execution exceeds timeout {0} to change state "
                       "to SUCCESS. Execution: {1}".format(timeout, ex))

                raise exceptions.TimeoutException(msg)

            ex = self.mistral_admin('execution-get', params=exec_id)
            exec_state = self.get_field_value(ex, 'State')

            if exec_state not in expected_states:
                msg = ("Execution state %s is not in expected "
                       "states: %s" % (exec_state, expected_states))

                raise exceptions.TempestException(msg)

            time.sleep(2)

        return True



# Copyright 2014 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
import json
import sys

import six

import mock

from mistralclient.api.v2 import action_executions as action_ex
from mistralclient.commands.v2 import action_executions as action_ex_cmd
from mistralclient.tests.unit import base

ACTION_EX_DICT = {
    'id': '123',
    'name': 'some',
    'workflow_name': 'thing',
    'workflow_namespace': '',
    'task_name': 'task1',
    'task_execution_id': "1-2-3-4",
    'state': 'RUNNING',
    'state_info': 'RUNNING somehow.',
    'accepted': True,
    'created_at': '1',
    'updated_at': '1',
}

ACTION_EX_RESULT = {"test": "is", "passed": "successfully"}
ACTION_EX_INPUT = {"param1": "val1", "param2": 2}

ACTION_EX_WITH_OUTPUT_DICT = ACTION_EX_DICT.copy()
ACTION_EX_WITH_OUTPUT_DICT.update({'output': json.dumps(ACTION_EX_RESULT)})

ACTION_EX_WITH_INPUT_DICT = ACTION_EX_DICT.copy()
ACTION_EX_WITH_INPUT_DICT.update({'input': json.dumps(ACTION_EX_INPUT)})

ACTION_EX = action_ex.ActionExecution(mock, ACTION_EX_DICT)
ACTION_EX_WITH_OUTPUT = action_ex.ActionExecution(
    mock,
    ACTION_EX_WITH_OUTPUT_DICT
)
ACTION_EX_WITH_INPUT = action_ex.ActionExecution(
    mock,
    ACTION_EX_WITH_INPUT_DICT
)


class TestCLIActionExecutions(base.BaseCommandTest):
    def test_create(self):
        (self.client.action_executions.create.
            return_value) = ACTION_EX_WITH_OUTPUT

        self.call(
            action_ex_cmd.Create,
            app_args=['some', '{"output": "Hello!"}']
        )

        self.assertDictEqual(
            ACTION_EX_RESULT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_create_save_result(self):
        (self.client.action_executions.create.
            return_value) = ACTION_EX_WITH_OUTPUT

        result = self.call(
            action_ex_cmd.Create,
            app_args=[
                'some', '{"output": "Hello!"}', '--save-result'
            ]
        )

        self.assertEqual(
            ('123', 'some', 'thing', '', 'task1', '1-2-3-4', 'RUNNING',
             'RUNNING somehow.', True, '1', '1'),
            result[1]
        )

    def test_create_run_sync(self):
        (self.client.action_executions.create.
            return_value) = ACTION_EX_WITH_OUTPUT

        self.call(
            action_ex_cmd.Create,
            app_args=[
                'some', '{"output": "Hello!"}', '--run-sync'
            ]
        )

        self.assertDictEqual(
            ACTION_EX_RESULT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_create_run_sync_and_save_result(self):
        (self.client.action_executions.create.
            return_value) = ACTION_EX_WITH_OUTPUT

        self.call(
            action_ex_cmd.Create,
            app_args=[
                'some', '{"output": "Hello!"}', '--save-result', '--run-sync'
            ]
        )

        self.assertDictEqual(
            ACTION_EX_RESULT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_update(self):
        states = ['IDLE', 'RUNNING', 'SUCCESS', 'ERROR', 'CANCELLED']

        for state in states:
            action_ex_dict = copy.deepcopy(ACTION_EX_DICT)
            action_ex_dict['state'] = state
            action_ex_dict['state_info'] = 'testing update'
            action_ex_obj = action_ex.ActionExecution(mock, action_ex_dict)
            self.client.action_executions.update.return_value = action_ex_obj

            result = self.call(
                action_ex_cmd.Update,
                app_args=['id', '--state', state]
            )

            expected_result = (
                action_ex_dict['id'],
                action_ex_dict['name'],
                action_ex_dict['workflow_name'],
                action_ex_dict['workflow_namespace'],
                action_ex_dict['task_name'],
                action_ex_dict['task_execution_id'],
                action_ex_dict['state'],
                action_ex_dict['state_info'],
                action_ex_dict['accepted'],
                action_ex_dict['created_at'],
                action_ex_dict['updated_at']
            )

            self.assertEqual(expected_result, result[1])

    def test_update_invalid_state(self):
        states = ['PAUSED', 'WAITING', 'DELAYED']

        # Redirect the stderr so it doesn't show during tox
        _stderr = sys.stderr
        sys.stderr = six.StringIO()

        for state in states:
            self.assertRaises(
                SystemExit,
                self.call,
                action_ex_cmd.Update,
                app_args=['id', '--state', state]
            )

        # Stop the redirection
        print(sys.stderr.getvalue())
        sys.stderr = _stderr

    def test_list(self):
        self.client.action_executions.list.return_value = [ACTION_EX]

        result = self.call(action_ex_cmd.List)

        self.assertEqual(
            [('123', 'some', 'thing', '', 'task1', '1-2-3-4', 'RUNNING', True,
              '1', '1')],
            result[1]
        )

    def test_get(self):
        self.client.action_executions.get.return_value = ACTION_EX

        result = self.call(action_ex_cmd.Get, app_args=['id'])

        self.assertEqual(
            ('123', 'some', 'thing', '', 'task1', '1-2-3-4', 'RUNNING',
             'RUNNING somehow.', True, '1', '1'), result[1]
        )

    def test_get_output(self):
        self.client.action_executions.get.return_value = ACTION_EX_WITH_OUTPUT

        self.call(action_ex_cmd.GetOutput, app_args=['id'])

        self.assertDictEqual(
            ACTION_EX_RESULT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_get_input(self):
        self.client.action_executions.get.return_value = ACTION_EX_WITH_INPUT

        self.call(action_ex_cmd.GetInput, app_args=['id'])

        self.assertDictEqual(
            ACTION_EX_INPUT,
            json.loads(self.app.stdout.write.call_args[0][0])
        )

    def test_delete(self):
        self.call(action_ex_cmd.Delete, app_args=['id'])

        self.client.action_executions.delete.assert_called_once_with('id')

    def test_delete_with_multi_names(self):
        self.call(action_ex_cmd.Delete, app_args=['id1', 'id2'])

        self.assertEqual(2, self.client.action_executions.delete.call_count)
        self.assertEqual(
            [mock.call('id1'), mock.call('id2')],
            self.client.action_executions.delete.call_args_list
        )



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import json

from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils


def format_list(environment=None):
    columns = (
        'Name',
        'Description',
        'Scope',
        'Created at',
        'Updated at'
    )

    if environment:
        data = (
            environment.name,
            environment.description,
            environment.scope,
            environment.created_at,
        )

        if hasattr(environment, 'updated_at'):
            data += (environment.updated_at or '<none>',)
        else:
            data += (None,)

    else:
        data = (tuple('<none>' for _ in range(len(columns))),)

    return columns, data


def format(environment=None):
    columns = (
        'Name',
        'Description',
        'Variables',
        'Scope',
        'Created at',
        'Updated at'
    )

    if environment:
        data = (environment.name,)

        if hasattr(environment, 'description'):
            data += (environment.description or '<none>',)
        else:
            data += (None,)

        data += (
            json.dumps(environment.variables, indent=4),
            environment.scope,
            environment.created_at,
        )

        if hasattr(environment, 'updated_at'):
            data += (environment.updated_at or '<none>',)
        else:
            data += (None,)

    else:
        data = (tuple('' for _ in range(len(columns))),)

    return columns, data


class List(base.MistralLister):
    """List all environments."""

    def _get_format_function(self):
        return format_list

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        return mistral_client.environments.list()


class Get(command.ShowOne):
    """Show specific environment."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument(
            'environment',
            help='Environment name'
        )

        parser.add_argument(
            '--export',
            default=False,
            action='store_true',
            help='Export the environment suitable for import'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        environment = mistral_client.environments.get(parsed_args.environment)

        if parsed_args.export:
            columns = ('name',
                       'description',
                       'scope',
                       'variables')

            data = (environment.name,
                    environment.description,
                    environment.scope,
                    json.dumps(environment.variables))

            return columns, data

        return format(environment)


class Create(command.ShowOne):
    """Create new environment."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'file',
            type=argparse.FileType('r'),
            help='Environment configuration file in JSON or YAML'
        )

        return parser

    def take_action(self, parsed_args):
        data = utils.load_content(parsed_args.file.read())

        mistral_client = self.app.client_manager.workflow_engine
        environment = mistral_client.environments.create(**data)

        return format(environment)


class Delete(command.Command):
    """Delete environment."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'environment',
            nargs='+',
            help='Name of environment(s).'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        utils.do_action_on_many(
            lambda s: mistral_client.environments.delete(s),
            parsed_args.environment,
            "Request to delete environment %s has been accepted.",
            "Unable to delete the specified environment(s)."
        )


class Update(command.ShowOne):
    """Update environment."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'file',
            type=argparse.FileType('r'),
            help='Environment configuration file in JSON or YAML'
        )

        return parser

    def take_action(self, parsed_args):
        data = utils.load_content(parsed_args.file.read())

        mistral_client = self.app.client_manager.workflow_engine
        environment = mistral_client.environments.update(**data)

        return format(environment)



# Copyright 2016 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mistralclient.api.v2 import members
from mistralclient.commands.v2 import members as member_cmd
from mistralclient.tests.unit import base

MEMBER_DICT = {
    'id': '123',
    'resource_id': '456',
    'resource_type': 'workflow',
    'project_id': '1111',
    'member_id': '2222',
    'status': 'pending',
    'created_at': '1',
    'updated_at': '1'
}

MEMBER = members.Member(mock, MEMBER_DICT)


class TestCLIWorkflowMembers(base.BaseCommandTest):
    def test_create(self):
        self.client.members.create.return_value = MEMBER

        result = self.call(
            member_cmd.Create,
            app_args=[MEMBER_DICT['resource_id'], MEMBER_DICT['resource_type'],
                      MEMBER_DICT['member_id']]
        )

        self.assertEqual(
            ('456', 'workflow', '1111', '2222', 'pending', '1', '1'),
            result[1]
        )

    def test_update(self):
        self.client.members.update.return_value = MEMBER

        result = self.call(
            member_cmd.Update,
            app_args=[MEMBER_DICT['resource_id'], MEMBER_DICT['resource_type'],
                      '-m', MEMBER_DICT['member_id']]
        )

        self.assertEqual(
            ('456', 'workflow', '1111', '2222', 'pending', '1', '1'),
            result[1]
        )

    def test_list(self):
        self.client.members.list.return_value = [MEMBER]

        result = self.call(
            member_cmd.List,
            app_args=[MEMBER_DICT['resource_id'], MEMBER_DICT['resource_type']]
        )

        self.assertListEqual(
            [('456', 'workflow', '1111', '2222', 'pending', '1', '1')],
            result[1]
        )

    def test_get(self):
        self.client.members.get.return_value = MEMBER

        result = self.call(
            member_cmd.Get,
            app_args=[MEMBER_DICT['resource_id'], MEMBER_DICT['resource_type'],
                      '-m', MEMBER_DICT['member_id']]
        )

        self.assertEqual(
            ('456', 'workflow', '1111', '2222', 'pending', '1', '1'),
            result[1]
        )

    def test_delete(self):
        self.call(
            member_cmd.Delete,
            app_args=[MEMBER_DICT['resource_id'], MEMBER_DICT['resource_type'],
                      MEMBER_DICT['member_id']]
        )

        self.client.members.delete.assert_called_once_with(
            MEMBER_DICT['resource_id'],
            MEMBER_DICT['resource_type'],
            MEMBER_DICT['member_id']
        )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json
import logging
import os.path

from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils

LOG = logging.getLogger(__name__)


def format_list(execution=None):
    return format(execution, lister=True)


def format(execution=None, lister=False):
    columns = (
        'ID',
        'Workflow ID',
        'Workflow name',
        'Workflow namespace',
        'Description',
        'Task Execution ID',
        'Root Execution ID',
        'State',
        'State info',
        'Created at',
        'Updated at'
    )
    # TODO(nmakhotkin) Add parent task id when it's implemented in API.

    if execution:
        state_info = (execution.state_info if not lister
                      else base.cut(execution.state_info))

        data = (
            execution.id,
            execution.workflow_id,
            execution.workflow_name,
            execution.workflow_namespace,
            execution.description,
            execution.task_execution_id or '<none>',
            execution.root_execution_id or '<none>',
            execution.state,
            state_info,
            execution.created_at,
            execution.updated_at or '<none>'
        )
    else:
        data = (tuple('' for _ in range(len(columns))),)

    return columns, data


class List(base.MistralLister):
    """List all executions."""

    def _get_format_function(self):
        return format_list

    def get_parser(self, parsed_args):
        parser = super(List, self).get_parser(parsed_args)
        parser.add_argument(
            '--task',
            nargs='?',
            help="Parent task execution ID associated with workflow "
                 "execution list.",
        )
        parser.add_argument(
            '--marker',
            type=str,
            help='The last execution uuid of the previous page, displays list '
                 'of executions after "marker".',
            default='',
            nargs='?'
        )
        parser.add_argument(
            '--limit',
            type=int,
            help='Maximum number of executions to return in a single result. '
                 'limit is set to %s by default. Use --limit -1 to fetch the '
                 'full result set.' % base.DEFAULT_LIMIT,
            nargs='?'
        )
        parser.add_argument(
            '--sort_keys',
            help='Comma-separated list of sort keys to sort results by. '
                 'Default: created_at. '
                 'Example: mistral execution-list --sort_keys=id,description',
            default='created_at',
            nargs='?'
        )
        parser.add_argument(
            '--sort_dirs',
            help='Comma-separated list of sort directions. Default: asc. '
                 'Example: mistral execution-list --sort_keys=id,description '
                 '--sort_dirs=asc,desc',
            default='asc',
            nargs='?'
        )
        parser.add_argument(
            '--filter',
            dest='filters',
            action='append',
            help='Filters. Can be repeated.'
        )

        return parser

    def _get_resources(self, parsed_args):
        if parsed_args.limit is None:
            parsed_args.limit = base.DEFAULT_LIMIT

            LOG.info(
                "limit is set to %s by default. Set "
                "the limit explicitly using \'--limit\', if required. "
                "Use \'--limit\' -1 to fetch the full result set.",
                base.DEFAULT_LIMIT
            )

        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.executions.list(
            task=parsed_args.task,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            sort_keys=parsed_args.sort_keys,
            sort_dirs=parsed_args.sort_dirs,
            **base.get_filters(parsed_args)
        )


class Get(command.ShowOne):
    """Show specific execution."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('execution', help='Execution identifier')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        execution = mistral_client.executions.get(parsed_args.execution)

        return format(execution)


class Create(command.ShowOne):
    """Create new execution."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'workflow_identifier',
            nargs='?',
            help='Workflow ID or name. Workflow name will be deprecated since '
                 'Mitaka.'
        )
        parser.add_argument(
            '--namespace',
            nargs='?',
            default='',
            help="Workflow namespace."
        )
        parser.add_argument(
            'workflow_input',
            nargs='?',
            help='Workflow input'
        )
        parser.add_argument(
            'params',
            nargs='?',
            help='Workflow additional parameters'
        )
        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            default='',
            help='Execution description'
        )
        parser.add_argument(
            '-s',
            dest='source_execution_id',
            nargs='?',
            help="Workflow Execution id which will allow operators to create "
                 "a new workflow execution based on the previously successful "
                 "executed workflow. Example: mistral execution-create -s "
                 "123e4567-e89b-12d3-a456-426655440000")

        return parser

    def take_action(self, parsed_args):
        if parsed_args.workflow_input:
            wf_input = utils.load_json(parsed_args.workflow_input)
        else:
            wf_input = {}

        if parsed_args.params:
            params = utils.load_json(parsed_args.params)
        else:
            params = {}

        mistral_client = self.app.client_manager.workflow_engine

        execution = mistral_client.executions.create(
            parsed_args.workflow_identifier,
            parsed_args.namespace,
            wf_input,
            parsed_args.description,
            parsed_args.source_execution_id,
            **params
        )

        return format(execution)


class Delete(command.Command):
    """Delete execution."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'execution',
            nargs='+',
            help='Id of execution identifier(s).'
        )

        parser.add_argument(
            '--force',
            default=False,
            action='store_true',
            help='Force the deletion of an execution. Might cause a cascade '
                 ' of errors if used for running executions.'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        force = parsed_args.force
        utils.do_action_on_many(
            lambda s: mistral_client.executions.delete(s, force=force),
            parsed_args.execution,
            "Request to delete execution %s has been accepted.",
            "Unable to delete the specified execution(s)."
        )


class Update(command.ShowOne):
    """Update execution."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'id',
            help='Execution identifier'
        )

        parser.add_argument(
            '-s',
            '--state',
            dest='state',
            choices=['RUNNING', 'PAUSED', 'SUCCESS', 'ERROR', 'CANCELLED'],
            help='Execution state'
        )

        parser.add_argument(
            '-e',
            '--env',
            dest='env',
            help='Environment variables'
        )

        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            help='Execution description'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        env = (
            utils.load_file(parsed_args.env)
            if parsed_args.env and os.path.isfile(parsed_args.env)
            else utils.load_content(parsed_args.env)
        )

        execution = mistral_client.executions.update(
            parsed_args.id,
            parsed_args.state,
            description=parsed_args.description,
            env=env
        )

        return format(execution)


class GetInput(command.Command):
    """Show execution input data."""

    def get_parser(self, prog_name):
        parser = super(GetInput, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution ID')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        ex_input = mistral_client.executions.get(parsed_args.id).input

        try:
            ex_input = json.loads(ex_input)
            ex_input = json.dumps(ex_input, indent=4) + "\n"
        except Exception:
            LOG.debug("Execution input is not JSON.")

        self.app.stdout.write(ex_input or "\n")


class GetOutput(command.Command):
    """Show execution output data."""

    def get_parser(self, prog_name):
        parser = super(GetOutput, self).get_parser(prog_name)

        parser.add_argument('id', help='Execution ID')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        output = mistral_client.executions.get(parsed_args.id).output

        try:
            output = json.loads(output)
            output = json.dumps(output, indent=4) + "\n"
        except Exception:
            LOG.debug("Execution output is not JSON.")

        self.app.stdout.write(output or "\n")



# Copyright 2016 - Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistralclient.api import base


class Member(base.Resource):
    resource_name = 'Member'


class MemberManager(base.ResourceManager):
    resource_class = Member

    def create(self, resource_id, resource_type, member_id):
        self._ensure_not_empty(
            resource_id=resource_id,
            resource_type=resource_type,
            member_id=member_id
        )

        data = {
            'member_id': member_id,
        }

        url = '/%ss/%s/members' % (resource_type, resource_id)

        return self._create(url, data)

    def update(self, resource_id, resource_type, member_id='',
               status='accepted'):
        if not member_id:
            member_id = self.http_client.project_id

        url = '/%ss/%s/members/%s' % (resource_type, resource_id, member_id)

        return self._update(url, {'status': status})

    def list(self, resource_id, resource_type):
        url = '/%ss/%s/members' % (resource_type, resource_id)

        return self._list(url, response_key='members')

    def get(self, resource_id, resource_type, member_id=None):
        self._ensure_not_empty(
            resource_id=resource_id,
            resource_type=resource_type,
        )

        if not member_id:
            member_id = self.http_client.project_id

        url = '/%ss/%s/members/%s' % (resource_type, resource_id, member_id)

        return self._get(url)

    def delete(self, resource_id, resource_type, member_id):
        self._ensure_not_empty(
            resource_id=resource_id,
            resource_type=resource_type,
            member_id=member_id
        )

        url = '/%ss/%s/members/%s' % (resource_type, resource_id, member_id)

        self._delete(url)



# Copyright 2014 - Mirantis, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import argparse

from osc_lib.command import command

from mistralclient.commands.v2 import base
from mistralclient import utils


def format_list(action=None):
    return format(action, lister=True)


def format(action=None, lister=False):
    columns = (
        'ID',
        'Name',
        'Is system',
        'Input',
        'Description',
        'Tags',
        'Created at',
        'Updated at'
    )

    if action:
        tags = getattr(action, 'tags', None) or []
        input = action.input if not lister else base.cut(action.input)
        desc = (action.description if not lister
                else base.cut(action.description))

        data = (
            action.id,
            action.name,
            action.is_system,
            input,
            desc,
            base.wrap(', '.join(tags)) or '<none>',
            action.created_at,
        )

        if hasattr(action, 'updated_at'):
            data += (action.updated_at,)
        else:
            data += (None,)
    else:
        data = (tuple('' for _ in range(len(columns))),)

    return columns, data


class List(base.MistralLister):
    """List all actions."""

    def _get_format_function(self):
        return format_list

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)

        parser.add_argument(
            '--filter',
            dest='filters',
            action='append',
            help='Filters. Can be repeated.'
        )

        return parser

    def _get_resources(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.actions.list(
            **base.get_filters(parsed_args)
        )


class Get(command.ShowOne):
    """Show specific action."""

    def get_parser(self, prog_name):
        parser = super(Get, self).get_parser(prog_name)

        parser.add_argument('action', help='Action (name or ID)')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        action = mistral_client.actions.get(parsed_args.action)

        return format(action)


class Create(base.MistralLister):
    """Create new action."""

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Action definition file'
        )
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag action will be marked as "public".'
        )

        return parser

    def _validate_parsed_args(self, parsed_args):
        if not parsed_args.definition:
            raise RuntimeError("Provide action definition file.")

    def _get_format_function(self):
        return format_list

    def _get_resources(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.actions.create(
            parsed_args.definition.read(),
            scope=scope
        )


class Delete(command.Command):
    """Delete action."""

    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            'action',
            nargs='+',
            help='Name or ID of action(s).'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        utils.do_action_on_many(
            lambda s: mistral_client.actions.delete(s),
            parsed_args.action,
            "Request to delete action %s has been accepted.",
            "Unable to delete the specified action(s)."
        )


class Update(base.MistralLister):
    """Update action."""

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='Action definition file'
        )
        parser.add_argument('--id', help='Action ID.')
        parser.add_argument(
            '--public',
            action='store_true',
            help='With this flag action will be marked as "public".'
        )

        return parser

    def _get_format_function(self):
        return format_list

    def _get_resources(self, parsed_args):
        scope = 'public' if parsed_args.public else 'private'

        mistral_client = self.app.client_manager.workflow_engine

        return mistral_client.actions.update(
            parsed_args.definition.read(),
            scope=scope,
            id=parsed_args.id
        )


class GetDefinition(command.Command):
    """Show action definition."""

    def get_parser(self, prog_name):
        parser = super(GetDefinition, self).get_parser(prog_name)

        parser.add_argument('name', help='Action name')

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine
        definition = mistral_client.actions.get(parsed_args.name).definition

        self.app.stdout.write(definition or "\n")


class Validate(command.ShowOne):
    """Validate action."""

    def _format(self, result=None):
        columns = ('Valid', 'Error')

        if result:
            data = (result.get('valid'), result.get('error'))
        else:
            data = (tuple('<none>' for _ in range(len(columns))),)

        return columns, data

    def get_parser(self, prog_name):
        parser = super(Validate, self).get_parser(prog_name)

        parser.add_argument(
            'definition',
            type=argparse.FileType('r'),
            help='action definition file'
        )

        return parser

    def take_action(self, parsed_args):
        mistral_client = self.app.client_manager.workflow_engine

        result = mistral_client.actions.validate(
            parsed_args.definition.read()
        )

        return self._format(result)
